@@ -14,8 +14,12 @@
 (*                                                                        *)
 (**************************************************************************)
 
-module From = Ast_405
-module To = Ast_404
+module From = Ast_406
+module To = Ast_405
+
+let migration_error loc missing_feature =
+  Location.raise_errorf ~loc
+    "migration error: %s is not supported before OCaml 4.06" missing_feature
 
 let rec copy_expression : From.Parsetree.expression -> To.Parsetree.expression =
  fun {
@@ -100,7 +104,7 @@ and copy_expression_desc :
       To.Parsetree.Pexp_coerce
         (copy_expression x0, copy_option copy_core_type x1, copy_core_type x2)
   | From.Parsetree.Pexp_send (x0, x1) ->
-      To.Parsetree.Pexp_send (copy_expression x0, x1.From.Asttypes.txt)
+      To.Parsetree.Pexp_send (copy_expression x0, copy_loc (fun x -> x) x1)
   | From.Parsetree.Pexp_new x0 ->
       To.Parsetree.Pexp_new (copy_loc copy_longident x0)
   | From.Parsetree.Pexp_setinstvar (x0, x1) ->
@@ -126,7 +130,7 @@ and copy_expression_desc :
   | From.Parsetree.Pexp_object x0 ->
       To.Parsetree.Pexp_object (copy_class_structure x0)
   | From.Parsetree.Pexp_newtype (x0, x1) ->
-      To.Parsetree.Pexp_newtype (x0.From.Asttypes.txt, copy_expression x1)
+      To.Parsetree.Pexp_newtype (copy_loc (fun x -> x) x0, copy_expression x1)
   | From.Parsetree.Pexp_pack x0 -> To.Parsetree.Pexp_pack (copy_module_expr x0)
   | From.Parsetree.Pexp_open (x0, x1, x2) ->
       To.Parsetree.Pexp_open
@@ -246,9 +250,13 @@ and copy_core_type_desc :
   | From.Parsetree.Ptyp_object (x0, x1) ->
       To.Parsetree.Ptyp_object
         ( List.map
-            (fun x ->
-              let x0, x1, x2 = x in
-              (x0.From.Asttypes.txt, copy_attributes x1, copy_core_type x2))
+            (function
+              | From.Parsetree.Otag (x0, x1, x2) ->
+                  ( copy_loc (fun x -> x) x0,
+                    copy_attributes x1,
+                    copy_core_type x2 )
+              | From.Parsetree.Oinherit _ ->
+                  migration_error Location.none "inheritance in object type")
             x0,
           copy_closed_flag x1 )
   | From.Parsetree.Ptyp_class (x0, x1) ->
@@ -263,7 +271,7 @@ and copy_core_type_desc :
           copy_option (fun x -> List.map copy_label x) x2 )
   | From.Parsetree.Ptyp_poly (x0, x1) ->
       To.Parsetree.Ptyp_poly
-        (List.map (fun x -> x.From.Asttypes.txt) x0, copy_core_type x1)
+        (List.map (fun x -> copy_loc (fun x -> x) x) x0, copy_core_type x1)
   | From.Parsetree.Ptyp_package x0 ->
       To.Parsetree.Ptyp_package (copy_package_type x0)
   | From.Parsetree.Ptyp_extension x0 ->
@@ -284,7 +292,7 @@ and copy_row_field : From.Parsetree.row_field -> To.Parsetree.row_field =
   function
   | From.Parsetree.Rtag (x0, x1, x2, x3) ->
       To.Parsetree.Rtag
-        ( copy_label x0,
+        ( copy_label x0.txt,
           copy_attributes x1,
           copy_bool x2,
           List.map copy_core_type x3 )
@@ -399,6 +407,8 @@ and copy_class_expr_desc :
       To.Parsetree.Pcl_constraint (copy_class_expr x0, copy_class_type x1)
   | From.Parsetree.Pcl_extension x0 ->
       To.Parsetree.Pcl_extension (copy_extension x0)
+  | From.Parsetree.Pcl_open (_, loc, _) ->
+      migration_error loc.Location.loc "module open in class expression"
 
 and copy_class_structure :
     From.Parsetree.class_structure -> To.Parsetree.class_structure =
@@ -426,7 +436,7 @@ and copy_class_field_desc :
       To.Parsetree.Pcf_inherit
         ( copy_override_flag x0,
           copy_class_expr x1,
-          copy_option (fun x -> x.From.Asttypes.txt) x2 )
+          copy_option (copy_loc (fun x -> x)) x2 )
   | From.Parsetree.Pcf_val x0 ->
       To.Parsetree.Pcf_val
         (let x0, x1, x2 = x0 in
@@ -545,11 +555,14 @@ and copy_with_constraint :
   | From.Parsetree.Pwith_module (x0, x1) ->
       To.Parsetree.Pwith_module
         (copy_loc copy_longident x0, copy_loc copy_longident x1)
-  | From.Parsetree.Pwith_typesubst x0 ->
+  | From.Parsetree.Pwith_typesubst ({ txt = Longident.Lident _; _ }, x0) ->
       To.Parsetree.Pwith_typesubst (copy_type_declaration x0)
-  | From.Parsetree.Pwith_modsubst (x0, x1) ->
-      To.Parsetree.Pwith_modsubst
-        (copy_loc (fun x -> x) x0, copy_loc copy_longident x1)
+  | From.Parsetree.Pwith_modsubst ({ txt = Longident.Lident x0; loc }, x1) ->
+      To.Parsetree.Pwith_modsubst ({ txt = x0; loc }, copy_loc copy_longident x1)
+  | From.Parsetree.Pwith_typesubst ({ loc; _ }, _x0) ->
+      migration_error loc "type substitution inside a submodule"
+  | From.Parsetree.Pwith_modsubst ({ loc; _ }, _x1) ->
+      migration_error loc "module substitution inside a submodule"
 
 and copy_signature : From.Parsetree.signature -> To.Parsetree.signature =
  fun x -> List.map copy_signature_item x
@@ -626,6 +639,8 @@ and copy_class_type_desc :
         (copy_arg_label x0, copy_core_type x1, copy_class_type x2)
   | From.Parsetree.Pcty_extension x0 ->
       To.Parsetree.Pcty_extension (copy_extension x0)
+  | From.Parsetree.Pcty_open (_, loc, _) ->
+      migration_error loc.Location.loc "module open in class type"
 
 and copy_class_signature :
     From.Parsetree.class_signature -> To.Parsetree.class_signature =
@@ -656,14 +671,14 @@ and copy_class_type_field_desc :
   | From.Parsetree.Pctf_val x0 ->
       To.Parsetree.Pctf_val
         (let x0, x1, x2, x3 = x0 in
-         ( x0.From.Asttypes.txt,
+         ( copy_loc (fun x -> x) x0,
            copy_mutable_flag x1,
            copy_virtual_flag x2,
            copy_core_type x3 ))
   | From.Parsetree.Pctf_method x0 ->
       To.Parsetree.Pctf_method
         (let x0, x1, x2, x3 = x0 in
-         ( x0.From.Asttypes.txt,
+         ( copy_loc (fun x -> x) x0,
            copy_private_flag x1,
            copy_virtual_flag x2,
            copy_core_type x3 ))
@@ -986,7 +1001,6 @@ and copy_loc :
   { To.Asttypes.txt = f0 txt; To.Asttypes.loc = copy_location loc }
 
 and copy_location : Location.t -> Location.t = fun x -> x
-
 and copy_bool : bool -> bool = function false -> false | true -> true
 
 let rec copy_toplevel_phrase :
@@ -1006,9 +1020,6 @@ and copy_directive_argument :
   | From.Parsetree.Pdir_bool x0 -> To.Parsetree.Pdir_bool (copy_bool x0)
 
 let copy_cases x = List.map copy_case x
-
 let copy_pat = copy_pattern
-
 let copy_expr = copy_expression
-
 let copy_typ = copy_core_type
