@@ -3,49 +3,41 @@ open! Import
 class map =
   object
     inherit Ppxlib_traverse_builtins.map
-
     inherit Ast.map
   end
 
 class iter =
   object
     inherit Ppxlib_traverse_builtins.iter
-
     inherit Ast.iter
   end
 
 class ['acc] fold =
   object
     inherit ['acc] Ppxlib_traverse_builtins.fold
-
     inherit ['acc] Ast.fold
   end
 
 class ['acc] fold_map =
   object
     inherit ['acc] Ppxlib_traverse_builtins.fold_map
-
     inherit ['acc] Ast.fold_map
   end
 
 class ['ctx] map_with_context =
   object
     inherit ['ctx] Ppxlib_traverse_builtins.map_with_context
-
     inherit ['ctx] Ast.map_with_context
   end
 
 class virtual ['res] lift =
   object
     inherit ['res] Ppxlib_traverse_builtins.lift
-
     inherit ['res] Ast.lift
   end
 
 let module_name = function None -> "_" | Some name -> name
-
 let enter name path = if String.is_empty path then name else path ^ "." ^ name
-
 let enter_opt name_opt path = enter (module_name name_opt) path
 
 class map_with_path =
@@ -102,8 +94,26 @@ class map_with_expansion_context =
   object (self)
     inherit [Expansion_context.Base.t] map_with_context as super
 
-    method! expression ctxt expr =
-      super#expression (Expansion_context.Base.enter_expr ctxt) expr
+    method! expression ctxt
+        { pexp_desc; pexp_loc; pexp_loc_stack; pexp_attributes } =
+      let ctxt = Expansion_context.Base.enter_expr ctxt in
+      let pexp_desc =
+        match pexp_desc with
+        | Pexp_letmodule (name, module_expr, body) ->
+            let name = self#loc (self#option self#string) ctxt name in
+            let module_expr =
+              self#module_expr
+                (ec_enter_module_opt ~loc:module_expr.pmod_loc name.txt ctxt)
+                module_expr
+            in
+            let body = self#expression ctxt body in
+            Pexp_letmodule (name, module_expr, body)
+        | _ -> self#expression_desc ctxt pexp_desc
+      in
+      let pexp_loc = self#location ctxt pexp_loc in
+      let pexp_loc_stack = self#list self#location ctxt pexp_loc_stack in
+      let pexp_attributes = self#attributes ctxt pexp_attributes in
+      { pexp_desc; pexp_loc; pexp_loc_stack; pexp_attributes }
 
     method! module_binding ctxt mb =
       super#module_binding
@@ -129,11 +139,10 @@ class map_with_expansion_context =
 
     method! value_binding ctxt { pvb_pat; pvb_expr; pvb_attributes; pvb_loc } =
       let all_var_names = var_names_of#pattern pvb_pat [] in
-      let var_name = Stdppx.List.last all_var_names in
       let in_binding_ctxt =
-        match var_name with
-        | None -> ctxt
-        | Some var_name ->
+        match all_var_names with
+        | [] | _ :: _ :: _ -> ctxt
+        | [ var_name ] ->
             Expansion_context.Base.enter_value ~loc:pvb_loc var_name ctxt
       in
       let pvb_pat = self#pattern ctxt pvb_pat in
@@ -146,31 +155,18 @@ class map_with_expansion_context =
 class sexp_of =
   object
     inherit [Sexp.t] Ast.lift
-
     method int = sexp_of_int
-
     method string = sexp_of_string
-
     method bool = sexp_of_bool
-
     method char = sexp_of_char
-
     method float = sexp_of_float
-
     method int32 = sexp_of_int32
-
     method int64 = sexp_of_int64
-
     method nativeint = sexp_of_nativeint
-
     method unit = sexp_of_unit
-
     method option = sexp_of_option
-
     method list = sexp_of_list
-
     method array : 'a. ('a -> Sexp.t) -> 'a array -> Sexp.t = sexp_of_array
-
     method other : 'a. 'a -> Sexp.t = fun _ -> Sexp.Atom "_"
 
     method record fields =
