@@ -48,59 +48,32 @@ module Context : sig
     | Object_type_field : object_field t
 
   val label_declaration : label_declaration t
-
   val constructor_declaration : constructor_declaration t
-
   val type_declaration : type_declaration t
-
   val type_extension : type_extension t
-
   val type_exception : type_exception t
-
   val extension_constructor : extension_constructor t
-
   val pattern : pattern t
-
   val core_type : core_type t
-
   val expression : expression t
-
   val value_description : value_description t
-
   val class_type : class_type t
-
   val class_type_field : class_type_field t
-
   val class_infos : _ class_infos t
-
   val class_expr : class_expr t
-
   val class_field : class_field t
-
   val module_type : module_type t
-
   val module_declaration : module_declaration t
-
   val module_type_declaration : module_type_declaration t
-
   val open_description : open_description t
-
   val include_infos : _ include_infos t
-
   val module_expr : module_expr t
-
   val value_binding : value_binding t
-
   val module_binding : module_binding t
-
   val pstr_eval : structure_item t
-
   val pstr_extension : structure_item t
-
   val psig_extension : signature_item t
-
   val rtag : row_field t
-
   val object_type_field : object_field t
 end
 
@@ -155,22 +128,39 @@ val declare_with_name_loc :
     attribute. *)
 
 val name : _ t -> string
-
 val context : ('a, _) t -> 'a Context.t
 
+val get_res :
+  ('a, 'b) t ->
+  ?mark_as_seen:bool (** default [true] *) ->
+  'a ->
+  ('b option, Location.Error.t NonEmptyList.t) result
+(** Gets the associated attribute value. Marks the attribute as seen unless
+    [mark_as_seen=false]. Returns an [Error] if the attribute is duplicated *)
+
 val get :
   ('a, 'b) t -> ?mark_as_seen:bool (** default [true] *) -> 'a -> 'b option
-(** Gets the associated attribute value. Marks the attribute as seen unless
-    [mark_as_seen=false]. *)
+(** See {!get_res}. Raises a located error if the attribute is duplicated *)
 
-val consume : ('a, 'b) t -> 'a -> ('a * 'b) option
-(** [consume t x] returns the value associated to attribute [t] on [x] if
+val consume_res :
+  ('a, 'b) t -> 'a -> (('a * 'b) option, Location.Error.t NonEmptyList.t) result
+(** [consume_res t x] returns the value associated to attribute [t] on [x] if
     present as well as [x] with [t] removed. *)
 
-val remove_seen : 'a Context.t -> packed list -> 'a -> 'a
+val consume : ('a, 'b) t -> 'a -> ('a * 'b) option
+(** See {!consume_res}. Raises a located exception in case of error. *)
+
+val remove_seen_res :
+  'a Context.t ->
+  packed list ->
+  'a ->
+  ('a, Location.Error.t NonEmptyList.t) result
 (** [remove_seen x attrs] removes the set of attributes matched by elements of
     [attrs]. Only remove them if they where seen by {!get} or {!consume}. *)
 
+val remove_seen : 'a Context.t -> packed list -> 'a -> 'a
+(** See {!remove_seen_res}. Raises in case of error. *)
+
 module Floating : sig
   type ('context, 'payload) t
 
@@ -182,11 +172,8 @@ module Floating : sig
       | Class_type_field : class_type_field t
 
     val structure_item : structure_item t
-
     val signature_item : signature_item t
-
     val class_field : class_field t
-
     val class_type_field : class_type_field t
   end
 
@@ -199,6 +186,9 @@ module Floating : sig
 
   val name : _ t -> string
 
+  val convert_res :
+    ('a, 'b) t list -> 'a -> ('b option, Location.Error.t NonEmptyList.t) result
+
   val convert : ('a, 'b) t list -> 'a -> 'b option
 end
 
@@ -207,17 +197,22 @@ val explicitly_drop : Ast_traverse.iter
     object. All attributes inside will be marked as handled. *)
 
 val check_unused : Ast_traverse.iter
-(** Raise if there are unused attributes *)
+(** Raise if there are unused attributes. *)
+
+val collect_unused_attributes_errors : Location.Error.t list Ast_traverse.fold
+(** Collect all errors due to unused attributes. *)
 
 val collect : Ast_traverse.iter
 (** Collect all attribute names. To be used in conjunction with
     {!check_all_seen}. *)
 
+val collect_unseen_errors : unit -> Location.Error.t list
+
 val check_all_seen : unit -> unit
 (** Check that all attributes collected by {!freshen_and_collect} have been:
 
     - matched at least once by one of: {!get}, {!consume} or {!Floating.convert}
-    - seen by [check_unused] (to allow white-listed attributed to pass through)
+    - seen by [check_unused] (to allow allowlisted attributed to pass through)
 
     This helps with faulty ppx rewriters that silently drop attributes. *)
 
@@ -231,10 +226,17 @@ val dropped_so_far_structure : structure -> string Loc.t list
     debug extensions that drop attributes. *)
 
 val dropped_so_far_signature : signature -> string Loc.t list
-
 val reset_checks : unit -> unit
 
 val pattern :
   ('a, 'b) t ->
   ('a, 'c, 'd) Ast_pattern.t ->
   ('a, 'b option -> 'c, 'd) Ast_pattern.t
+
+val pattern_res :
+  ('a, 'b) t ->
+  ('a, 'c, 'd) Ast_pattern.t ->
+  ( 'a,
+    'b option -> 'c,
+    ('d, Location.Error.t NonEmptyList.t) result )
+  Ast_pattern.t
