@@ -22,7 +22,6 @@ module Make (M : sig
   type t
 
   val get_loc : t -> Location.t
-
   val end_marker : (t, unit) Attribute.Floating.t
 
   module Transform (T : T1) : sig
@@ -34,9 +33,7 @@ module Make (M : sig
   end
 
   val parse : Lexing.lexbuf -> t list
-
   val pp : Format.formatter -> t -> unit
-
   val to_sexp : t -> Sexp.t
 end) =
 struct
@@ -46,12 +43,15 @@ struct
           let loc =
             { Location.loc_start = pos; loc_end = pos; loc_ghost = false }
           in
-          Location.raise_errorf ~loc "ppxlib: [@@@@@@%s] attribute missing"
-            (Attribute.Floating.name M.end_marker)
+          Error
+            ( Location.Error.createf ~loc "ppxlib: [@@@@@@%s] attribute missing"
+                (Attribute.Floating.name M.end_marker),
+              [] )
       | x :: l -> (
-          match Attribute.Floating.convert [ M.end_marker ] x with
-          | None -> loop (x :: acc) l
-          | Some () -> (List.rev acc, (M.get_loc x).loc_start)
+          match Attribute.Floating.convert_res [ M.end_marker ] x with
+          | Ok None -> loop (x :: acc) l
+          | Ok (Some ()) -> Ok (List.rev acc, (M.get_loc x).loc_start)
+          | Error e -> Error e
           | exception Failure _ -> loop (x :: acc) l)
     in
     loop [] l
@@ -59,9 +59,7 @@ struct
   let remove_loc =
     object
       inherit Ast_traverse.map
-
       method! location _ = Location.none
-
       method! location_stack _ = []
     end
 
@@ -70,7 +68,6 @@ struct
   end)
 
   let remove_loc x = M_map.apply remove_loc x
-
   let rec last prev = function [] -> prev | x :: l -> last x l
 
   let diff_asts ~generated ~round_trip =
@@ -144,7 +141,8 @@ struct
         match_loop ~end_pos ~mismatch_handler ~expected ~source
 
   let do_match ~pos ~expected ~mismatch_handler source =
-    let source, end_pos = extract_prefix ~pos source in
+    let open Result in
+    extract_prefix ~pos source >>| fun (source, end_pos) ->
     match_loop ~end_pos ~mismatch_handler ~expected ~source
 end
 
@@ -153,7 +151,6 @@ module Str = Make (struct
   type t = structure_item
 
   let get_loc x = x.pstr_loc
-
   let end_marker = end_marker_str
 
   module Transform (T : T1) = struct
@@ -161,9 +158,7 @@ module Str = Make (struct
   end
 
   let parse = Parse.implementation
-
   let pp = Pprintast.structure_item
-
   let to_sexp = Ast_traverse.sexp_of#structure_item
 end)
 
@@ -172,7 +167,6 @@ module Sig = Make (struct
   type t = signature_item
 
   let get_loc x = x.psig_loc
-
   let end_marker = end_marker_sig
 
   module Transform (T : T1) = struct
@@ -180,14 +174,20 @@ module Sig = Make (struct
   end
 
   let parse = Parse.interface
-
   let pp = Pprintast.signature_item
-
   let to_sexp = Ast_traverse.sexp_of#signature_item
 end)
 
 (*$*)
 
-let match_structure = Str.do_match
+let match_structure_res = Str.do_match
 
-let match_signature = Sig.do_match
+let match_structure ~pos ~expected ~mismatch_handler l =
+  match_structure_res ~pos ~expected ~mismatch_handler l
+  |> Result.handle_error ~f:(fun (err, _) -> Location.Error.raise err)
+
+let match_signature_res = Sig.do_match
+
+let match_signature ~pos ~expected ~mismatch_handler l =
+  match_signature_res ~pos ~expected ~mismatch_handler l
+  |> Result.handle_error ~f:(fun (err, _) -> Location.Error.raise err)
