@@ -4,6 +4,8 @@ type t = {
   file_path : string;
   main_module_name : string;
   submodule_path : string loc list;
+  enclosing_module : string;
+  enclosing_value : string option;
   value : string loc option;
   in_expr : bool;
 }
@@ -17,13 +19,16 @@ let top_level ~file_path =
     file_path;
     main_module_name;
     submodule_path = [];
+    enclosing_module = main_module_name;
+    enclosing_value = None;
     value = None;
     in_expr = false;
   }
 
 let file_path t = t.file_path
-
 let main_module_name t = t.main_module_name
+let enclosing_module t = t.enclosing_module
+let enclosing_value t = t.enclosing_value
 
 let submodule_path t =
   List.rev_map ~f:(fun located -> located.txt) t.submodule_path
@@ -41,15 +46,24 @@ let fully_qualified_path t =
 let enter_expr t = { t with in_expr = true }
 
 let enter_module ~loc module_name t =
-  if t.in_expr then t
+  if t.in_expr then { t with enclosing_module = module_name }
   else
-    { t with submodule_path = { txt = module_name; loc } :: t.submodule_path }
+    {
+      t with
+      submodule_path = { txt = module_name; loc } :: t.submodule_path;
+      enclosing_module = module_name;
+    }
 
 let enter_value ~loc value_name t =
-  if t.in_expr then t else { t with value = Some { txt = value_name; loc } }
+  if t.in_expr then { t with enclosing_value = Some value_name }
+  else
+    {
+      t with
+      value = Some { txt = value_name; loc };
+      enclosing_value = Some value_name;
+    }
 
 let to_string_path t = String.concat ~sep:"." (t.file_path :: submodule_path t)
-
 let with_string_path f ~loc ~path = f ~loc ~path:(to_string_path path);;
 
 let module M = struct
