@@ -1,9 +1,11 @@
 open! Import
 
 val lident : string -> Longident.t
-
 val core_type_of_type_declaration : type_declaration -> core_type
 
+val name_type_params_in_td_res :
+  type_declaration -> (type_declaration, Location.Error.t NonEmptyList.t) result
+
 val name_type_params_in_td : type_declaration -> type_declaration
 
 val combinator_type_of_type_declaration :
@@ -15,14 +17,19 @@ val gen_symbol : ?prefix:string -> unit -> string
     @param prefix default = "_x" *)
 
 val string_of_core_type : core_type -> string
-
 val assert_no_attributes : attributes -> unit
-
 val assert_no_attributes_in : Ast_traverse.iter
+val attributes_errors : attributes -> Location.Error.t list
+val collect_attributes_errors : Location.Error.t list Ast_traverse.fold
+
+val get_type_param_name_res :
+  core_type * (variance * injectivity) ->
+  (string Loc.t, Location.Error.t NonEmptyList.t) result
+(** [get_type_param_name_res tp] returns the string identifier associated with
+    [tp] if it is a type parameter, as a result. *)
 
 val get_type_param_name : core_type * (variance * injectivity) -> string Loc.t
-(** [get_tparam_id tp] returns the string identifier associated with [tp] if it
-    is a type parameter. *)
+(** See {!get_type_param_name_res}. Raises a located error in case of failure. *)
 
 (** [(new type_is_recursive rec_flag tds)#go ()] returns whether [rec_flag, tds]
     is really a recursive type. We disregard recursive occurrences appearing in
@@ -33,11 +40,8 @@ class type_is_recursive :
   -> type_declaration list
   -> object
        inherit Ast_traverse.iter
-
        val type_names : string list
-
        method return_true : unit -> unit
-
        method go : unit -> rec_flag
      end
 
@@ -45,9 +49,7 @@ val really_recursive : rec_flag -> type_declaration list -> rec_flag
 (** [really_recursive rec_flag tds = (new type_is_recursive rec_flag tds)#go ()] *)
 
 val loc_of_payload : attribute -> Location.t
-
 val loc_of_attribute : attribute -> Location.t
-
 val loc_of_extension : extension -> Location.t
 
 val curry_applications : expression -> expression
