@@ -41,7 +41,6 @@ module Pattern = struct
     { name; dot_suffixes = String.Set.of_list (dot_suffixes name) }
 
   let name t = t.name
-
   let matches t matched = String.Set.mem matched t.dot_suffixes
 end
 
@@ -50,8 +49,8 @@ let get_outer_namespace name =
   | None -> None
   | Some i -> Some (String.sub name ~pos:0 ~len:i)
 
-module Whitelisted = struct
-  (* White list the following attributes, as well as all their dot suffixes.
+module Allowlisted = struct
+  (* Allow list the following attributes, as well as all their dot suffixes.
 
      Since these attributes are interpreted by the compiler itself, we cannot check
      at the level of a ppx rewriter that they have been properly interpreted, so
@@ -93,9 +92,11 @@ module Whitelisted = struct
         "ocaml.warn_on_literal_pattern";
         "ocaml.warnerror";
         "ocaml.warning";
+        "ocaml.toplevel_printer" (*Interpreted by the toplevel/utop*);
+        "toplevel_printer" (*Interpreted by the toplevel/utop*);
       ]
 
-  (* White list the following extensions.
+  (* Allow list the following extensions.
 
      Since these extensions are interpreted by the compiler itself, we cannot check
      at the level of a ppx rewriter that they have been properly interpreted, so
@@ -103,29 +104,22 @@ module Whitelisted = struct
   *)
   let extensions = create_set [ "ocaml.error"; "ocaml.extension_constructor" ]
 
-  let is_whitelisted ~kind name =
+  let is_allowlisted ~kind name =
     match kind with
     | `Attribute -> String.Set.mem name attributes
     | `Extension -> String.Set.mem name extensions
 
   let get_attribute_list () = String.Set.elements attributes
-
   let get_extension_list () = String.Set.elements extensions
 end
 
 module Reserved_namespaces = struct
   let tbl : (string, unit) Hashtbl.t = Hashtbl.create 16
-
   let reserve ns = Hashtbl.add_exn tbl ~key:ns ~data:()
-
   let () = reserve "merlin"
-
   let () = reserve "reason"
-
   let () = reserve "refmt"
-
   let () = reserve "metaocaml"
-
   let () = reserve "ocamlformat"
 
   let is_in_reserved_namespaces name =
@@ -136,8 +130,8 @@ module Reserved_namespaces = struct
   let check_not_reserved ~kind name =
     let kind, list =
       match kind with
-      | `Attribute -> ("attribute", Whitelisted.attributes)
-      | `Extension -> ("extension", Whitelisted.extensions)
+      | `Attribute -> ("attribute", Allowlisted.attributes)
+      | `Extension -> ("extension", Allowlisted.extensions)
     in
     if String.Set.mem name list then
       Printf.ksprintf failwith
@@ -157,7 +151,6 @@ let ignore_checks name =
 
 module Registrar = struct
   type element = { fully_qualified_name : string; declared_at : Caller_id.t }
-
   type all_for_context = { mutable all : element String.Map.t }
 
   type 'a t = {
@@ -215,12 +208,12 @@ module Registrar = struct
       fold_dot_suffixes name ~init:all.all ~f:(fun name acc ->
           String.Map.add name t acc)
 
-  let spellcheck t context ?(white_list = []) name =
+  let spellcheck t context ?(allowlist = []) name =
     let all =
       let all = get_all_for_context t context in
       String.Map.fold (fun key _ acc -> key :: acc) all.all []
     in
-    match Spellcheck.spellcheck (all @ white_list) name with
+    match Spellcheck.spellcheck (all @ allowlist) name with
     | Some _ as x -> x
     | None -> (
         let other_contexts =
@@ -267,13 +260,23 @@ module Registrar = struct
                       Format.fprintf ppf ",@ "))
                  others pp_text last current_context))
 
-  (* TODO: hint spelling errors regarding reserved namespaces names and white
-     listed names instead of taking an optional [white_list] parameter. *)
-  let raise_errorf t context ?white_list fmt (name : string Loc.t) =
-    Printf.ksprintf
-      (fun msg ->
-        match spellcheck t context name.txt ?white_list with
-        | None -> Location.raise_errorf ~loc:name.loc "%s" msg
-        | Some s -> Location.raise_errorf ~loc:name.loc "%s.\n%s" msg s)
-      fmt name.txt
+  module Error = struct
+    (* TODO: hint spelling errors regarding reserved namespaces names and allowlisted
+       names instead of taking an optional [allowlist] parameter. *)
+    let createf t context ?allowlist fmt (name : string Loc.t) =
+      Printf.ksprintf
+        (fun msg ->
+          match spellcheck t context name.txt ?allowlist with
+          | None -> Location.Error.createf ~loc:name.loc "%s" msg
+          | Some s -> Location.Error.createf ~loc:name.loc "%s.\n%s" msg s)
+        fmt name.txt
+
+    let raise_errorf t context ?allowlist fmt (name : string Loc.t) =
+      Location.Error.raise @@ createf t context ?allowlist fmt name
+
+    let error_extensionf t context ?allowlist fmt (name : string Loc.t) =
+      Location.Error.to_extension @@ createf t context ?allowlist fmt name
+  end
+
+  let raise_errorf = Error.raise_errorf
 end
