@@ -9,50 +9,31 @@ module type Comparisons = sig
   type t
 
   val compare : t -> t -> int
-
   val equal : t -> t -> bool
-
   val ( = ) : t -> t -> bool
-
   val ( < ) : t -> t -> bool
-
   val ( > ) : t -> t -> bool
-
   val ( <> ) : t -> t -> bool
-
   val ( <= ) : t -> t -> bool
-
   val ( >= ) : t -> t -> bool
-
   val min : t -> t -> t
-
   val max : t -> t -> t
 end
 
 module Poly = struct
   let compare = compare
-
   let equal = ( = )
-
   let ( = ) = ( = )
-
   let ( < ) = ( < )
-
   let ( > ) = ( > )
-
   let ( <> ) = ( <> )
-
   let ( <= ) = ( <= )
-
   let ( >= ) = ( >= )
-
   let min = min
-
   let max = max
 end
 
 include (Poly : Comparisons with type t := int)
-
 module Array = Array
 
 module Bool = struct
@@ -72,7 +53,6 @@ end
 
 module Char = struct
   include Char
-
   include (Poly : Comparisons with type t := char)
 end
 
@@ -174,21 +154,22 @@ end
 
 module Int = struct
   let max_int = max_int
-
   let to_string = string_of_int
 
   include (Poly : Comparisons with type t := int)
 end
 
+module Either = struct
+  type ('a, 'b) t = Left of 'a | Right of 'b
+end
+
 module List = struct
   include List
 
   include struct
     (* shadow non-tail-recursive functions *)
     let merge = `not_tail_recursive
-
     let remove_assoc = `not_tail_recursive
-
     let remove_assq = `not_tail_recursive
 
     let rev_mapi list ~f =
@@ -203,14 +184,12 @@ module List = struct
       fold_left2 (rev list1) (rev list2) ~init ~f:(fun acc x y -> f x y acc)
 
     let map list ~f = rev (rev_map list ~f)
-
     let mapi list ~f = rev (rev_mapi list ~f)
 
     let fold_right list ~init ~f =
       fold_left (List.rev list) ~init ~f:(fun acc x -> f x acc)
 
     let append x y = rev_append (rev x) y
-
     let concat list = fold_right list ~init:[] ~f:append
 
     let rev_combine list1 list2 =
@@ -226,6 +205,16 @@ module List = struct
       rev (fold_left2 list1 list2 ~init:[] ~f:(fun acc x y -> f x y :: acc))
   end
 
+  let partition_map p l =
+    let rec part left right = function
+      | [] -> (rev left, rev right)
+      | x :: l -> (
+          match p x with
+          | Either.Left v -> part (v :: left) right l
+          | Either.Right v -> part left (v :: right) l)
+    in
+    part [] [] l
+
   let init ~len ~f =
     let rec loop ~len ~pos ~f ~acc =
       if pos >= len then List.rev acc
@@ -240,9 +229,7 @@ module List = struct
         match option with None -> tail | Some head -> head :: tail)
 
   let filter_opt list = rev (rev_filter_opt list)
-
   let filter_map list ~f = rev_filter_opt (rev_map list ~f)
-
   let concat_map list ~f = concat (map list ~f)
 
   let rec find_map list ~f =
@@ -292,12 +279,28 @@ end
 
 module Option = struct
   let is_some = function None -> false | Some _ -> true
-
   let iter t ~f = match t with None -> () | Some x -> f x
-
   let map t ~f = match t with None -> None | Some x -> Some (f x)
-
   let value t ~default = match t with None -> default | Some x -> x
+  let to_list t = match t with None -> [] | Some x -> [ x ]
+end
+
+module Result = struct
+  let bind t ~f = match t with Ok a -> f a | Error e -> Error e
+  let map t ~f = match t with Ok a -> Ok (f a) | Error e -> Error e
+  let map_error t ~f = match t with Ok a -> Ok (f a) | Error e -> Error e
+  let ( >>= ) t f = bind t ~f
+  let ( >>| ) t f = map t ~f
+  let handle_error t ~f = match t with Ok a -> a | Error e -> f e
+end
+
+module NonEmptyList = struct
+  type 'a t = 'a * 'a list
+
+  let ( @ ) (t1, q1) (t2, q2) = (t1, q1 @ (t2 :: q2))
+  let hd = fst
+  let to_list (t, q) = t :: q
+  let map ~f (t, q) = (f t, List.map ~f q)
 end
 
 module Out_channel = struct
@@ -321,13 +324,9 @@ module String = struct
   include String
 
   let is_empty (t : t) = length t = 0
-
   let prefix t len = sub t ~pos:0 ~len
-
   let suffix t len = sub t ~pos:(length t - len) ~len
-
   let drop_prefix t len = sub t ~pos:len ~len:(length t - len)
-
   let drop_suffix t len = sub t ~pos:0 ~len:(length t - len)
 
   let is_prefix t ~prefix =
@@ -383,11 +382,8 @@ module String = struct
         Some (sub t ~pos:0 ~len:i, sub t ~pos:(i + 1) ~len:(length t - i - 1))
 
   let capitalize_ascii = Stdlib.String.capitalize_ascii
-
   let lowercase_ascii = Stdlib.String.lowercase_ascii
-
   let uncapitalize_ascii = Stdlib.String.uncapitalize_ascii
-
   let split_on_char t ~sep = Stdlib.String.split_on_char sep t
 
   include (Poly : Comparisons with type t := string)
@@ -403,7 +399,5 @@ module String = struct
 end
 
 let ( @ ) = List.append
-
 let output oc bytes ~pos ~len = output oc bytes pos len
-
 let output_substring oc string ~pos ~len = output_substring oc string pos len
